"""Experiment CLI."""

import json
import time
from datetime import datetime
from enum import Enum
from typing import Annotated, Optional

import jwt
import typer
import yaml
from pytz import UTC
from requests.exceptions import HTTPError
from rich.table import Table
from slices_bi_client.models.update_resources_request import UpdateResourcesRequest
from slices_expauth_client.exceptions import ApiException
from slices_expauth_client.models.experiment import Experiment
from slices_expauth_client.models.experiment_create import ExperimentCreate
from slices_expauth_client.models.experiment_resource import ExperimentResource
from slices_expauth_client.models.experiment_update import ExperimentUpdate
from slices_expauth_client.models.resource_state import ResourceState
from slices_projauth_client.exceptions import NotFoundException as ProjectNotFoundException

from . import console, err_console
from .context import Context
from .util.pagination import get_next_page_marker
from .util.textual_duration import (
    STR_TO_TIMEDELTA_HELP,
    duration_or_datetime,
    str_to_timedelta,
    validate_timedelta_str,
)
from .utils import datetime_with_local_timezone, friendly_datetime, get_error_detail, is_experiment_id, is_project_id

experiment_app = typer.Typer(short_help="Experiments", no_args_is_help=True)


@experiment_app.command(name="create", rich_help_panel="Commands")
def create_experiment(
    ctx: Context,
    name: Annotated[str, typer.Option("--name", "-n", help="Name of the experiment.")],
    expiration: Annotated[datetime, typer.Option("--expiration", "--exp", "-e")],
    description: Annotated[Optional[str], typer.Option("--description", "--desc", "-d")] = None,
    project_id: Annotated[Optional[str], typer.Option("--project-id", "-p")] = None,
):
    """Create a new experiment."""

    if project_id:
        if not is_project_id(project_id):
            project = ctx.obj.api.get_project_by_name(project_id)

            if not project:
                err_console.print(f"[red]Could not find project with name {project_id} :magnifying_glass_tilted_left:")
                raise typer.Exit(1)

            project_id = project.id

        project_token = ctx.obj.api.get_project_jwt(project_id)
    else:
        project_token = ctx.obj.get_current_project_token()

        if not project_token:
            err_console.print(
                "[bold red]No default project set. "
                "Please specify a default project with `slices project use <name>`, or pass a project id. :x:"
            )
            raise typer.Exit(1)

    try:
        experiment = ctx.obj.api.experiments_api.create_experiment(
            ExperimentCreate(
                friendly_name=name,
                description=description,
                expires_at=datetime_with_local_timezone(expiration),
                project_token=project_token,
            )
        )

    except ApiException as err:
        err_console.print("[bold red]Could not create experiment :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err

    if not console.is_terminal:
        console.print(experiment.id)
        return

    console.print(
        f"[green]:sparkles: Created experiment [bold]{experiment.friendly_name}[/bold] "
        f"that expires on [bold]{friendly_datetime(experiment.expires_at)}[/]",
    )
    console.print(f"Experiment ID: {experiment.id}")


@experiment_app.command(name="list", rich_help_panel="Commands")
def list_experiments(  # noqa: C901
    ctx: Context,
    include_expired: Annotated[
        Optional[bool], typer.Option("--include-expired", "-e", help="Also show expired experiments.")
    ] = False,
    include_archived: Annotated[
        Optional[bool], typer.Option("--include-archived", "-A", help="Also show archived experiments.")
    ] = False,
    all: Annotated[Optional[bool], typer.Option("--all", "-a", help="Show all experiments.")] = None,
    all_projects: Annotated[
        Optional[bool], typer.Option("--all-projects", "-P", help="Show experiments from all your projects.")
    ] = None,
    show_ids: Annotated[bool, typer.Option("--show-ids", "-I", help="Show internal IDs.")] = False,
    name: Annotated[Optional[str], typer.Option("--name", "-n", help="Name of the experiment to list.")] = None,
):
    """List experiments."""

    if all is True:
        include_expired = True
        include_archived = True

    project_id = ctx.obj.get_current_project_id()
    if all_projects:
        project_id = None

    project_id_to_name_map = {}
    if project_id is None:
        projects = ctx.obj.api.projects_api.get_projects()
        project_id_to_name_map = {project.id: project.name for project in projects.projects}

    experiments = ctx.obj.api.experiments_api.get_experiments(
        include_expired=include_expired,
        include_archived=include_archived,
        project_id=project_id,
        friendly_name=name,
    )

    projects = ctx.obj.api.projects_api.get_projects()

    columns = ["Name", "Description", "Created at", "Expires at"]
    if include_archived:
        columns.append("Archived")

    if project_id is None:
        columns.insert(1, "Project")
    if show_ids:
        columns.insert(0, "ID")

    title = "Experiments"
    if project_id:
        project = ctx.obj.api.projects_api.get_project(project_id)
        title = f"Experiments in project [bold]{project.name}[/]"

    table = Table(*columns, title=title)

    for exp in experiments.experiments:
        exp_expires_at = friendly_datetime(exp.expires_at)
        if exp.expires_at < datetime.now(UTC):
            exp_expires_at = f"[red]{exp_expires_at}"

        row = [
            exp.friendly_name,
            exp.description,
            friendly_datetime(exp.created_at),
            exp_expires_at,
        ]
        if include_archived:
            row.append("Yes" if exp.archived else "No")
        if not project_id:
            row.insert(1, project_id_to_name_map.get(exp.project_id, exp.project_id))

        if show_ids:
            row.insert(0, exp.id)

        table.add_row(*row, style="red" if exp.archived else "")

    console.print(table)


def _get_experiment(ctx, experiment_id_or_name: str):
    if is_experiment_id(experiment_id_or_name):
        return ctx.obj.api.experiments_api.get_experiment(experiment_id=experiment_id_or_name)

    experiments = ctx.obj.api.experiments_api.get_experiments(
        project_id=ctx.obj.require_project_id(),
        friendly_name=experiment_id_or_name,
    )

    if not len(experiments.experiments) > 0:
        err_console.print(
            f"[red]Could not find experiment with name {experiment_id_or_name} :magnifying_glass_tilted_left:"
        )
        raise typer.Exit(1)

    return experiments.experiments[0]


@experiment_app.command("show", rich_help_panel="Commands")
def show_experiment(
    ctx: Context,
    name: Annotated[str, typer.Argument(..., help="Name or ID of experiment to show.")],
    json: Annotated[bool, typer.Option("--json", "-j", is_flag=True, help="Output as JSON.")] = False,
):
    """Show experiment info."""
    experiment = _get_experiment(ctx, name)

    if json:
        console.print_json(experiment.model_dump_json())
        return

    project = None
    try:
        project = ctx.obj.api.projects_api.get_project(experiment.project_id)
    except ProjectNotFoundException:
        err_console.print("[orange]Could not fetch project info :warning:")

    table = Table(box=None)
    table.add_column("Name", justify="right")
    table.add_column("Value")
    table.add_row("ID", experiment.id)
    table.add_row("Name", experiment.friendly_name)
    table.add_row("Project ID", experiment.project_id)
    table.add_row("Project name", project.name if project else "** Unknown **")
    table.add_row("Created on", friendly_datetime(experiment.created_at))
    table.add_row("Expires on", friendly_datetime(experiment.expires_at))
    table.add_row("Deleted", "Yes" if experiment.archived else "No")

    console.print(table)


@experiment_app.command("list-resources", rich_help_panel="Commands")
def list_experiment_resources(
    ctx: Context,
    name: Annotated[str, typer.Argument(..., help="Name or ID of experiment to show.")],
    include_expired: Annotated[bool, typer.Option()] = True,
    include_terminated: Annotated[bool, typer.Option()] = True,
    show_ids: Annotated[bool, typer.Option("--show-ids", "-I", help="Show internal IDs.")] = False,
):
    """Show experiment resources."""

    experiment = _get_experiment(ctx, name)

    resources_page = ctx.obj.api.experiments_api.get_experiment_resources(
        experiment_id=experiment.id,
        include_expired=include_expired,
        include_terminated=include_terminated,
    )

    cols = [
        "Friendly Name",
        "Site",
        "Type",
        "Status",
        "Description",
        "Created At",
        "Expires At",
        "Terminated At",
    ]
    if show_ids:
        cols.insert(0, "ID")

    table = Table(
        *cols,
        title=f"Resources of experiment {experiment.friendly_name}",
    )

    while resources_page:
        for resource in resources_page.experiment_resources:
            row = [
                resource.friendly_name,
                resource.site_id,
                resource.type,
                resource.status,
                resource.description,
                friendly_datetime(resource.created_at),
                friendly_datetime(resource.expires_at) if resource.expires_at else "",
                friendly_datetime(resource.terminated_at) if resource.terminated_at else "",
            ]
            if show_ids:
                row.insert(0, resource.id)

            table.add_row(*row)
        marker = get_next_page_marker(resources_page.experiment_resource_links)
        resources_page = (
            ctx.obj.api.experiments_api.get_experiment_resources(
                experiment_id=experiment.id,
                include_expired=include_expired,
                include_terminated=include_terminated,
                marker=marker,
            )
            if marker
            else None
        )

    console.print(table)


class PrintResourcesFormat(Enum):
    """Output format for Resources."""

    text = "text"
    json = "json"
    yaml = "yaml"


@experiment_app.command("print-resources", rich_help_panel="Commands")
def print_experiment_resources(
    ctx: Context,
    name: Annotated[str, typer.Argument(..., help="Name or ID of experiment to show.")],
    include_expired: Annotated[bool, typer.Option()] = True,
    include_terminated: Annotated[bool, typer.Option()] = True,
    format: Annotated[PrintResourcesFormat, typer.Option(case_sensitive=False)] = "yaml",  # pyright: ignore [reportArgumentType]  weird but this works
):
    """Print experiment resources."""

    experiment = _get_experiment(ctx, name)

    resources_page = ctx.obj.api.experiments_api.get_experiment_resources(
        experiment_id=experiment.id,
        include_expired=include_expired,
        include_terminated=include_terminated,
    )

    res = []

    def _resource_to_dict(resource: ExperimentResource) -> dict:
        return {
            "id": resource.id,
            "friendly_name": resource.friendly_name,
            "service_id": resource.service_id,
            "site_id": resource.site_id,
            "type": resource.type,
            "status": resource.status.name,
            "description": resource.description,
            "created_at": resource.created_at.astimezone().isoformat(),
            "expires_at": resource.expires_at.astimezone().isoformat() if resource.expires_at else None,
            "terminated_at": resource.terminated_at.astimezone().isoformat() if resource.terminated_at else None,
            "exp_name": experiment.friendly_name,
            "exp_id": experiment.id,
        }

    while resources_page:
        for resource in resources_page.experiment_resources:
            res.append(_resource_to_dict(resource))  # noqa: PERF401
        marker = get_next_page_marker(resources_page.experiment_resource_links)
        resources_page = (
            ctx.obj.api.experiments_api.get_experiment_resources(
                experiment_id=experiment.id,
                include_expired=include_expired,
                include_terminated=include_terminated,
                marker=marker,
            )
            if marker
            else None
        )

    if format == PrintResourcesFormat.text:
        for r in res:
            for k, v in r.items():  # pyright: ignore [reportAttributeAccessIssue]
                typer.echo(f"{k}={v}")
            typer.echo("\n")
    elif format == PrintResourcesFormat.json:
        console.print(json.dumps(res), no_wrap=True, soft_wrap=True)
    elif format == PrintResourcesFormat.yaml:
        console.print(yaml.safe_dump(res))
    else:
        raise ValueError(f"Unimplemented format requested: {format}")


def _renew_experiment_resources(ctx: Context, experiment: Experiment):
    resources_page = ctx.obj.api.experiments_api.get_experiment_resources(
        experiment_id=experiment.id,
    )

    # Get all BI sites in which this experiment has resources.
    # Other types of resources can also be parsed here.
    bi_sites_with_experiment_resources = set()
    while resources_page:
        bi_sites_with_experiment_resources.update(
            {
                resource.site_id
                for resource in resources_page.experiment_resources
                if resource.service_id == "bi" and resource.site_id
            }
        )
        marker = get_next_page_marker(resources_page.experiment_resource_links)
        resources_page = (
            ctx.obj.api.experiments_api.get_experiment_resources(
                experiment_id=experiment.id,
                marker=marker,
            )
            if marker
            else None
        )

    experiment_token = ctx.obj.api.get_experiment_jwt(experiment.id)

    if not bi_sites_with_experiment_resources:
        console.print("[yellow]:information: No BI resources to renew.")

    renew_errors = []
    for site_id in bi_sites_with_experiment_resources:
        try:
            ctx.obj.api.get_bi_resources_api(site_id, experiment_token).update_resources(
                experiment.project_id,
                experiment.id,
                UpdateResourcesRequest(expires_at=experiment.expires_at),
            )

            console.print(f":recycle: Renewed resources on site {site_id}")
        except ApiException as err:  # noqa: PERF203
            err_console.print(f"[bold red]Could not renew resources on site {site_id} :x:")
            err_console.print(f"[red]{get_error_detail(err.body)}")
            renew_errors.append(site_id)

    if renew_errors:
        raise typer.Exit(1)


@experiment_app.command("renew", rich_help_panel="Commands")
def renew_experiment(
    ctx: Context,
    name: Annotated[str, typer.Argument(help="Name or ID of experiment that must be renewed.")],
    expiration: Annotated[
        datetime,
        typer.Argument(help="New expiration date or duration.", parser=duration_or_datetime),
    ],
    all_resources: Annotated[
        bool,
        typer.Option(
            "--all-resources/--skip-resources",
            "-A/-S",
            help="Renew all resources in the experiment to the same date.",
            is_flag=True,
        ),
    ] = True,
):
    """Extend the duration of an experiment."""

    if is_experiment_id(name):
        experiment_id = name
    else:
        experiments = ctx.obj.api.experiments_api.get_experiments(
            project_id=ctx.obj.require_project_id(),
            friendly_name=name,
        )

        if not len(experiments.experiments) > 0:
            err_console.print(f"[red]Could not find experiment with name {name} :magnifying_glass_tilted_left:")
            raise typer.Exit(1)

        experiment_id = experiments.experiments[0].id

    try:
        renewed_experiment = ctx.obj.api.experiments_api.update_experiment(
            experiment_id, ExperimentUpdate(expires_at=datetime_with_local_timezone(expiration))
        )
    except ApiException as err:
        err_console.print("[bold red]Could not renew experiment :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err

    console.print(
        f":recycle: Renewed experiment '{renewed_experiment.friendly_name}' "
        f"with ID '{renewed_experiment.id}' until {friendly_datetime(renewed_experiment.expires_at)}"
    )

    if not all_resources:
        console.print("[yellow]:warning: Resources were not renewed. :warning:")
    else:
        _renew_experiment_resources(ctx, renewed_experiment)


@experiment_app.command("delete", rich_help_panel="Commands")
def delete_experiment(
    ctx: Context,
    name: Annotated[str, typer.Argument(..., help="Name or ID of experiment that must be deleted.")],
    force: Annotated[bool, typer.Option("--force", "-f", help="Do not ask for confirmation", is_flag=True)] = False,
):
    """Delete an experiment by name or ID."""
    if is_experiment_id(name):
        experiment_id = name
    else:
        experiments = ctx.obj.api.experiments_api.get_experiments(
            project_id=ctx.obj.require_project_id(),
            friendly_name=name,
        )

        if not len(experiments.experiments) > 0:
            err_console.print(f"[red]Could not find experiment with name {name} :magnifying_glass_tilted_left:")
            raise typer.Exit(1)

        experiment_id = experiments.experiments[0].id

        if not force:
            typer.confirm(f"Are you sure you want to delete experiment {name} with ID '{experiment_id}'?", abort=True)

    try:
        deleted_experiment = ctx.obj.api.experiments_api.delete_experiment(experiment_id)
    except ApiException as err:
        err_console.print("[bold red]Could not delete experiment :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err

    console.print(
        f":wastebasket: Deleted experiment '{deleted_experiment.friendly_name}' with ID '{deleted_experiment.id}'"
    )


@experiment_app.command(
    "jwt",
    help="Get credentials for a specific experiment.",
    # hidden=True,
    rich_help_panel="Advanced",
)
def retrieve_jwt(
    ctx: Context,
    name: str,
    decode: Annotated[bool, typer.Option("--decode", "-d", help="Decode the token.")] = False,
):
    """Print experiment credentials."""
    experiment = _get_experiment(ctx, name)

    try:
        token = ctx.obj.api.get_experiment_jwt(experiment.id)

        if decode:
            console.print_json(jwt.decode(token, options={"verify_signature": False}))
        else:
            print(token)
    except HTTPError as e:
        try:
            error_response = e.response.json()
            err_console.print(
                f"[bold red]Error: Could not get token. :x:[/]\n[red]{error_response['error_description']}"
            )
        except:  # noqa: E722
            err_console.print("[bold red]Error: Could not get token. :x:")
            err_console.print(f"[red]{str(e.response.text)}")
        raise typer.Exit(1) from None


class WaitResourcesOutputFormat(Enum):
    """Output format of the --wait command."""

    console = "console"  # text output that overrides the previous line for each new status check
    text = "text"  # text output with a new line for each new status check
    json = "json"  # one json array per status check
    yaml = "yaml"  # one yaml fragment (multi-line) per status check


@experiment_app.command("wait", rich_help_panel="Commands")
def wait_for_resource_status(  # noqa: C901, PLR0912, PLR0915
    ctx: Context,
    experiment_name: Annotated[str, typer.Argument(help="Name or ID of experiment to show.")],
    resource_names: list[str] = typer.Argument(  # noqa B008: this is a typer hack, see next line
        # Reason for this typer syntax: see https://github.com/fastapi/typer/issues/518#issuecomment-1919910970
        default=None,
        help="Name or ID of the resources to wait for. May be empty to wait for all resources.",
    ),
    status: Annotated[
        ResourceState,
        typer.Option(case_sensitive=False, callback=lambda v: v.lower(), help="The status for which to wait."),
    ] = ResourceState.UP,
    timeout: Annotated[
        str,
        typer.Option(
            callback=validate_timedelta_str,
            help=f"How long needs to be waited for the resources to reach the requested status? "
            f"Format: {STR_TO_TIMEDELTA_HELP}",
        ),
    ] = "10 minutes",
    interval: Annotated[
        str,
        typer.Option(
            callback=validate_timedelta_str,
            help=f"Time between status checks. Format: {STR_TO_TIMEDELTA_HELP}",
        ),
    ] = "1s",
    format: Annotated[
        WaitResourcesOutputFormat,
        typer.Option(
            case_sensitive=False,
            help='The output format. "console" will print a table.'
            "If in a terminal, this table will be updated instead of printing a new table.",
        ),
    ] = "console",  # pyright: ignore [reportArgumentType]  weird but this works
    include_expired: Annotated[bool, typer.Option(help="Do not ignore EXPIRED Resources.")] = False,
    include_terminated: Annotated[bool, typer.Option(help="Do not ignore DELETED Resources.")] = False,
    ignore_missing: Annotated[
        bool,
        typer.Option(
            help="Do not fail if one of the specified resources does not exist."
            "This is useful if the resource will be created after starting the wait."
        ),
    ] = False,
):
    """Wait for the specified resources in the experiment to reach a requested status (default UP).

    If no resources are specified, wait for all resources in the experiment.
    Exit with exit status 1 if the resources can never reach the target state (= they become FAILED or DELETED).
    Exit with exit status 2 if the resources did not reach the target state before the timeout.
    """

    start = datetime.now(UTC)
    deadline = start + str_to_timedelta(timeout)
    interval_s = str_to_timedelta(interval).total_seconds()

    all_resources = resource_names is None

    experiment = _get_experiment(ctx, experiment_name)

    ready = False
    first = True

    in_alt_screen = False
    try:
        if format == WaitResourcesOutputFormat.console:
            in_alt_screen = console.set_alt_screen(enable=True)
        elif format == WaitResourcesOutputFormat.text:
            console.print(f"Waiting for {status.name!r}...")

        while not ready:
            if not first:
                if datetime.now(UTC) > deadline:
                    if in_alt_screen:
                        console.set_alt_screen(enable=False)
                        in_alt_screen = False
                    if all_resources:
                        rname = "All resources"
                    elif len(resource_names) == 1:
                        rname = f"Resource {resource_names[0]!r}"
                    else:
                        rname = "Resources"
                    err_console.print(
                        f"[red]{rname} did not become {status.name} before the requested timeout ({timeout}):x:"
                    )
                    raise typer.Exit(2)

                time.sleep(interval_s)

            resources_page = ctx.obj.api.experiments_api.get_experiment_resources(
                experiment_id=experiment.id,
                include_expired=include_expired,
                include_terminated=include_terminated,
            )

            new_ready = True
            res = []

            def _resource_to_dict(resource: ExperimentResource) -> dict:
                return {
                    "id": resource.id,
                    "friendly_name": resource.friendly_name,
                    "status": resource.status.name,
                }

            while resources_page:
                for resource in resources_page.experiment_resources:
                    if all_resources or resource.id in resource_names or resource.friendly_name in resource_names:
                        res.append(_resource_to_dict(resource))  # noqa: PERF401
                        new_ready = new_ready and resource.status == status
                        if (
                            status == ResourceState.FAILED
                            and status
                            not in (
                                ResourceState.FAILED,
                                ResourceState.DELETED,
                            )
                        ) or (status == ResourceState.DELETED and status != ResourceState.DELETED):
                            if in_alt_screen:
                                console.set_alt_screen(enable=False)
                                in_alt_screen = False
                            err_console.print(
                                f"[red]Resource {resource.friendly_name} has status {resource.status.name}: "
                                f"it can never become {status.name}. :x:"
                            )
                            raise typer.Exit(1)
                marker = get_next_page_marker(resources_page.experiment_resource_links)
                resources_page = (
                    ctx.obj.api.experiments_api.get_experiment_resources(
                        experiment_id=experiment.id,
                        include_expired=include_expired,
                        include_terminated=include_terminated,
                        marker=marker,
                    )
                    if marker
                    else None
                )

            saw_all = all_resources or len(resource_names) == len(res)
            if not saw_all and not ignore_missing:
                if in_alt_screen:
                    console.set_alt_screen(enable=False)
                    in_alt_screen = False
                seen_names = [r["friendly_name"] for r in res]
                seen_ids = [r["id"] for r in res]
                missing_resources = [
                    rname for rname in resource_names if rname not in seen_names and rname not in seen_ids
                ]
                err_console.print(f"[red]Not all requested resources seen, missing: {', '.join(missing_resources)} :x:")
                raise typer.Exit(1)

            now = datetime.now()
            if format == WaitResourcesOutputFormat.console:
                console.clear()  # will do nothing if not in a terminal
                table = Table(title=f"Status at {friendly_datetime(now, include_seconds=True)}")
                table.add_column("Resource", justify="right")
                table.add_column("Status", style="magenta", no_wrap=True)
                for r in res:
                    table.add_row(r["friendly_name"], r["status"])
                console.print(f"Waiting for {status.name!r}...")
                console.print(table)
            elif format == WaitResourcesOutputFormat.text:
                for r in res:
                    console.print(
                        f"{friendly_datetime(now, include_seconds=True)} "
                        f"- Resource {r['friendly_name']!r} - Status {r['status']!r}"
                    )
            elif format == WaitResourcesOutputFormat.json:
                console.print(json.dumps(res))
            elif format == WaitResourcesOutputFormat.yaml:
                console.print("---")
                console.print(yaml.safe_dump(res))
            else:
                raise ValueError(f"Unimplemented format requested: {format}")

            ready = new_ready
            first = False
    finally:
        if in_alt_screen:
            console.set_alt_screen(enable=False)

    if format in (WaitResourcesOutputFormat.console, WaitResourcesOutputFormat.text):
        console.print(f"Status {status.name!r} reached.")
